import re
import operator
import random

import pygame

import data
from world import *


# Targets are all objects subject to a rule. Eg,
# in the rule ALL WATER CLOSE TO DOG FREEZE, the
# target is "ALL WATER".
TARGETS = []

def EVERYTHING(objects):
    """
    Return all objects, plain and simple.
    """
    return objects
TARGETS.append((EVERYTHING, r"""(EVERYTHING|ANYTHING)"""))

def WHOEVER(objects):
    """
    Return all persons, but not other objects.
    """
    for object in objects:
        if isinstance(object, Person):
            yield object
TARGETS.append((WHOEVER, r"""(WHOEVER IS|EVERYBODY|ANYONE|PLAYERS|EVERYONE)"""))

def ALL_STAR(objects, what):
    """
    Return all objects from a given group.
    """
    for object in objects:
        if re.match(object.what, what):
            yield object
TARGETS.append((ALL_STAR, r"""ALL (?P<what>\w+)"""))

def STAR(objects, what):
    """
    Catch-all filter, for phrases omitting ALL.
    """
    for object in objects:
        if re.match(object.what, what):
            yield object
TARGETS.append((STAR, r"""(?P<what>\w+)"""))


# Filter are functions that limit the targets. In the 
# example ALL DOGS CLOSE TO PLAYERS FREEZE, the target
# is all dogs, but the filter will limit them to those
# close to player.
FILTERS = []

# A filter factory that compares distances.
def compare_distance(compare, distance, allow_self=False):
    def filter_(objects, world, groups, what):
        for object in objects:
            for item in world:
                if re.match(item.what, what):
                    # Get distance between two objects.
                    dx = abs(object.rect.x - item.rect.x)
                    dy = abs(object.rect.y - item.rect.y)
                    if (allow_self or (object != item)) and (
                            compare(dx, distance) and
                            compare(dy, distance)):
                        yield object
    return filter_

# Distance < 64.
CLOSE_TO_STAR = compare_distance(operator.lt, 64)
FILTERS.append((CLOSE_TO_STAR, r"""(CLOSE TO|NEAR) (?P<what>\w+)"""))

# Distance < 32.
TOUCHING_STAR = compare_distance(operator.lt, 32)
FILTERS.append((TOUCHING_STAR, r"""TOUCHING (?P<what>\w+)"""))

# Distance < 16.
OVER_STAR = compare_distance(operator.lt, 16)
FILTERS.append((OVER_STAR, r"""OVER (?P<what>\w+)"""))

# Distance > 64, for far objects. In this case we allow self
# in the comparison, but I don't remember why.
FAR_FROM_STAR = compare_distance(operator.gt, 64, allow_self=True)
FILTERS.append((FAR_FROM_STAR, r"""FAR FROM (?P<what>\w+)"""))

def NAMED(objects, world, groups, name):
    """
    Compare by name.
    """
    for object in objects:
        obj_name = getattr(object, 'name', 'Nameless')
        if re.match(obj_name, name):
            yield object
FILTERS.append((NAMED, r"""(NAMED|CALLED) (?P<name>\w+)"""))


# Actions are the fun part of the system. Here is where
# things happen.
ACTIONS = []

def WINS(objects, world, groups):
    """
    Sets a winner.
    """
    for object in objects:
        object.winner = True
ACTIONS.append((WINS, r"""WINS?"""))

def MOVE_DIRECTION(objects, world, groups, direction):
    """
    Move object in a given direction.
    """
    # Get the direction vector.
    dir = {'NORTH'    : ( 0, -1), 'N' : ( 0, -1),
           'SOUTH'    : ( 0,  1), 'S' : ( 0,  1),
           'EAST'     : ( 1,  0), 'E' : ( 1,  0),
           'WEST'     : (-1,  0), 'W' : (-1,  0),
           'NORTHWEST': (-1, -1), 'NW': (-1, -1),
           'NORTHEAST': ( 1, -1), 'NE': ( 1, -1),
           'SOUTHWEST': (-1,  1), 'SW': (-1,  1),
           'SOUTHEAST': ( 1,  1), 'SE': ( 1,  1),
           }[direction]
    for object in objects:
        # And set the direction in the object, so it will move.
        object.direction = dir

        # Add speed. 
        object._init_queue.insert(0,
                lambda s: setattr(s, 'speed', 3))
        # And cleanup.
        object._finish_queue.insert(0,
                lambda s, speed=object.speed: setattr(s, 'speed', speed))
ACTIONS.append((MOVE_DIRECTION,
    r"""MOVES? TO(WARDS)? (?P<direction>NORTHWEST|NORTHEAST|NORTH|SOUTHWEST|SOUTHEAST|SOUTH|EAST|WEST|NW|NE|N|SW|SE|S|W|E)"""))

def MOVE(objects, world, groups, speed):
    """
    Move object with a given speed.
    """
    for object in objects:
        sp = {'FAST': 8, 'FASTER': 8,
              'SUPER FAST': 12, 'SUPER FASTER': 12,
              'SLOW': 2, 'SLOWER': 2,
              'SUPER SLOW': 1, 'SUPER SLOWER': 1,
              None: 3}[speed]

        # Add speed. 
        object._init_queue.insert(0,
                lambda s, sp=sp: setattr(s, 'speed', sp))

        # Add random direction only to non-players.
        if not isinstance(object, FreeWilling):
            direction = (random.choice([-1,0,1]), random.choice([-1,0,1]))
            object._init_queue.insert(0,
                    lambda s, direction=direction: setattr(s, 'direction', direction))
            
        # And cleanup.
        object._finish_queue.insert(0,
                lambda s, speed=object.speed: setattr(s, 'speed', speed))
ACTIONS.append((MOVE, r"""MOVES?(\s(?P<speed>(SUPER )?(FAST(ER)?|SLOW(ER)?)))?"""))

def FLY(objects, world, groups):
    """
    Objects fly, avoiding collision with others.
    """
    for object in objects:
        # Remove all collidables from this object, so it doesn't hit anything.
        object._init_queue.insert(0, 
                lambda s: setattr(s, 'collidables', {}))
        # Also, remove object from its group, so nothing collides with it.
        object._init_queue.insert(0,
                lambda s: s.remove(groups[s.group]))
        # And restore everything.
        object._finish_queue.insert(0, 
                lambda s, c=object.collidables: setattr(s, 'collidables', c))
        object._finish_queue.insert(0,
                lambda s: s.add(groups[s.group]))
ACTIONS.append((FLY, r"""(FLIES|FLY)"""))

def WALK_OVER_WATER(objects, world, groups):
    """
    Name says all.
    """
    for object in objects:
        # Remove WATER from the collidables of the object.
        # First we make a copy.
        copy_ = object.collidables.copy()
        # We move slowly over water.
        copy_['WATER'] = 0.5
        # Add new collidables.
        object._init_queue.insert(0,
                lambda s: setattr(s, 'collidables', copy_))
        # And cleanup.
        object._finish_queue.insert(0,
                lambda s, c=object.collidables: setattr(s, 'collidables', c))
ACTIONS.append((WALK_OVER_WATER, r"""(WALKS? OVER WATER|CAN SWIM|SWIMS?)"""))

def REPRODUCE(objects, world, groups):
    """
    Objects reproduce into new ones.
    """
    for object in objects:
        # Put a low probability, otherwise we fill the screen with new objects.
        # Also, limit to 99 objects.
        if len(world) < 100 and random.random() < 0.005:
            # Create a copy of the object.
            new_ = object.copy()

            # And put it in the neighbourhood.
            width = new_.rect.width
            height = new_.rect.height
            new_.rect.x = random.choice([new_.rect.x - width, new_.rect.x + width])
            new_.rect.y = random.choice([new_.rect.y - height, new_.rect.y + height])

            # Add it to the sprite group.
            new_.add(groups.setdefault(object.group, pygame.sprite.Group()))

            # But we need to add in the proper place, so that it
            # doesnt overwrite the player/computer sprites.
            index = groups['ALL'].spritelist.index(object)
            groups['ALL'].add(new_, index) 

            # And add it to the world.
            world.append(new_)
ACTIONS.append((REPRODUCE, r"""(REPRODUCE|REPLICATE)"""))

def INVISIBLE(objects, world, groups):
    """
    Objects are invisible!
    """
    for object in objects:
        # Create invisible image from a single pixel.
        pixel = pygame.image.load(data.load("invisible_1x1.png")).convert_alpha()
        image = pygame.transform.scale(pixel, object.image.get_size())

        # And change the image...
        object._init_queue.insert(0,
                lambda s: setattr(s, 'image', image))
        # ...as well as the swapping images.
        object._init_queue.insert(0, 
                lambda s: setattr(s, '_images', []))
        # And cleanup.
        object._finish_queue.insert(0,
                lambda s, image=object.image: setattr(s, 'image', image))
        object._finish_queue.insert(0,
                lambda s, images=object._images: setattr(s, '_images', images))
ACTIONS.append((INVISIBLE, r"""(IS|ARE) INVISIBLE"""))

def DIE(objects, world, groups):
    """
    Objects die!
    """
    for object in objects:
        object.alive = False
        object.kill()
        i = world.index(object)
        del world[i]
ACTIONS.append((DIE, r"""(DIES|DIE|IS DEAD|ARE DEAD|DISAPPEAR|DISAPPEARS)"""))

def TELEPORT(objects, world, groups):
    """
    Teleport to a random location.
    """
    for object in objects:
        # Get new position.
        object.rect.x = random.randint(0, 640-32)
        object.rect.y = random.randint(0, 480-32)

        # Remove from allsprites to check collision,
        # but first get the index for ordering.
        index = groups['ALL'].spritelist.index(object)
        groups['ALL'].remove(object)
        # Check collision.
        while pygame.sprite.spritecollide(object, groups['ALL'], 0):
            object.rect.x = random.randint(0, 640-32)
            object.rect.y = random.randint(32, 480-64)

        # Now we put it back in the proper place.
        groups['ALL'].add(object, index)  
ACTIONS.append((TELEPORT, r"""(TELEPORT|APPARATE)"""))

def TRANSFORM_INTO_STAR(objects, world, groups, target):
    """
    Transform one object into another.
    """
    for object in objects:
        # Get original position.
        index = groups['ALL'].spritelist.index(object)

        # Create new thing and delete old one..
        klass = globals()[target.title().rstrip('s')]
        new_ = klass((object.rect.x, object.rect.y), object.speed, world)
        object.kill()
        i = world.index(object)
        del world[i]

        # Add to world and groups.
        world.append(new_)
        new_.add(groups.setdefault(new_.group, pygame.sprite.Group()))
        groups['ALL'].add(new_, index)
ACTIONS.append((TRANSFORM_INTO_STAR, r"""(TRANSFORMS? (IN|TO|INTO)|BECOMES?) (?P<target>\w+)"""))

def MELT(objects, world, groups):
    """
    Melt => transform into water.
    """
    for object in objects:
        # Get original position.
        index = groups['ALL'].spritelist.index(object)

        # Create new thing and delete old one..
        new_ = Water((object.rect.x, object.rect.y), object.speed, world)
        object.kill()
        i = world.index(object)
        del world[i]

        # Add to world and groups.
        world.append(new_)
        new_.add(groups.setdefault(new_.group, pygame.sprite.Group()))
        groups['ALL'].add(new_, index)
ACTIONS.append((MELT, r"""MELTS?"""))

def this_to_that(groups, what, group, image):
    """
    A function factory to transform between objects.

    This function just changes the image, group and "what",
    so it's not versatile as TRANSFORM_INTO_STAR. It's just used
    in FREEZE.
    """
    def func(obj, groups=groups, what=what, group=group, image=image):
        # Remove from the original group.
        obj.remove(groups[obj.group])
        # Set new vallues.
        obj.what = what
        obj.group = group
        if isinstance(image, basestring):
            image = pygame.image.load(data.load(image)).convert_alpha()
        obj.image = image
        # And add to new group.
        obj.add(groups.setdefault(group, pygame.sprite.Group()))
    return func

def FREEZE(objects, world, groups):
    """
    Freeze object.
    """
    for object in objects:
        # If this is water, we turn it to ice.
        if re.match(object.what, 'WATER'):
            # Use the this_to_that function, to preserve the object.
            object._init_queue.insert(0, this_to_that(groups, 'ICE', 'ICE', 'ice.png'))
            object._finish_queue.insert(0, this_to_that(groups, '(WATER|LAKE|POND)', 'WATER', object.image))
        else:
            # Otherwise, speed is set to zero.
            object._init_queue.insert(0,
                    lambda s: setattr(s, 'speed', 0))
            # And cleanup.
            object._finish_queue.insert(0,
                    lambda s, speed=object.speed: setattr(s, 'speed', speed))
ACTIONS.append((FREEZE, r"""FREEZES?"""))

def MUTE(objects, world, groups, m=None):
    """
    Sets mute attribute, making it impossible to add new rules.

    This flag is checked during the main loop.
    """
    for object in objects:
        # Set mute.
        object._init_queue.insert(0,
                lambda s: setattr(s, 'mute', True))
        # Unset.
        object._finish_queue.insert(0,
                lambda s: setattr(s, 'mute', False))
ACTIONS.append((MUTE, r"""(IS|ARE) MUTE"""))


def check_rule(rule):
    """
    Check if rule is understood.
    """
    for select, regexp in TARGETS:
        m = re.match(regexp, rule)
        if m:
            rule = rule[len(m.group()):].strip()
            break
    else:
        return False

    # Filter are optional.
    for filter_, regexp in FILTERS:
        m = re.match(regexp, rule)
        if m:
            rule = rule[len(m.group()):].strip()
            break

    for action, regexp in ACTIONS:
        m = re.match(regexp, rule)
        if m:
            rule = rule[len(m.group()):].strip()
            if not rule:
                return True

    return False


def apply_rule(rule, objects, groups):
    """
    Apply rule to group of objects.
    """
    for select, regexp in TARGETS:
        m = re.match(regexp, rule)
        if m:
            targets = select(objects, **m.groupdict())
            rule = rule[len(m.group()):].strip()
            break

    for filter_, regexp in FILTERS:
        m = re.match(regexp, rule)
        if m:
            targets = filter_(targets, objects, groups, **m.groupdict())
            rule = rule[len(m.group()):].strip()
            break

    for action, regexp in ACTIONS:
        m = re.match(regexp, rule)
        if m:
            action(targets, objects, groups, **m.groupdict())

